/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


/* list of all subsystems that support the shiny new error facility */
enum {SS_RECV, SS_NET, SS_ORTP_RECV, SS_AUDIOD, SS_EXEC, SS_CLOSE_ON_FORK, SS_SIGNAL,
	SS_STRING, SS_DAEMON, SS_STAT, SS_TIME, SS_GRAB_CLIENT, SS_HTTP_RECV,
	SS_RECV_COMMON, SS_FILTER_CHAIN, SS_WAV, SS_COMPRESS, SS_OGGDEC, SS_FILTER,
	SS_MP3DEC};
#define NUM_SS (SS_MP3DEC + 1)


#define NET_ERRORS \
	PARA_ERROR(NET, SEND, "send error"), \
	PARA_ERROR(NET, RECV, "receive error"), \
	PARA_ERROR(NET, SOCKET, "socket error"), \
	PARA_ERROR(NET, CONNECT, "connect error"), \
	PARA_ERROR(NET, ACCEPT, "accept error"), \
	PARA_ERROR(NET, SETSOCKOPT, "failed to set socket options"), \
	PARA_ERROR(NET, BIND, "bind error"), \
	PARA_ERROR(NET, NAME_TOO_LONG, "name too long for struct sockaddr_un"), \
	PARA_ERROR(NET, CHMOD, "failed to set socket mode"), \
	PARA_ERROR(NET, SENDMSG, "sendmsg() failed"), \
	PARA_ERROR(NET, RECVMSG, "recvmsg() failed"), \
	PARA_ERROR(NET, SCM_CREDENTIALS, "did not receive SCM credentials"), \
	PARA_ERROR(NET, LISTEN, "listen error"), \
	PARA_ERROR(NET, RECV_PATTERN, "did not receive expected pattern")


#define ORTP_RECV_ERRORS \
	PARA_ERROR(ORTP_RECV, MSG_TO_BUF, "failed to extract rtp packet"), \
	PARA_ERROR(ORTP_RECV, ORTP_SYNTAX, "ottp syntax error"), \
	PARA_ERROR(ORTP_RECV, TOO_MANY_BAD_CHUNKS, "too many consecutive bad chunks"), \
	PARA_ERROR(ORTP_RECV, SHORT_HEADER, "header packet too short to contain a valid header"), \
	PARA_ERROR(ORTP_RECV, INVALID_HEADER, "invalid header packet"), \
	PARA_ERROR(ORTP_RECV, OVERRUN, "outout buffer overrun"), \


#define HTTP_RECV_ERRORS \
	PARA_ERROR(HTTP_RECV, SEND_HTTP_REQUEST, "failed to send http request"), \
	PARA_ERROR(HTTP_RECV, MISSING_OK, "did not receive OK message from peer"), \
	PARA_ERROR(HTTP_RECV, HTTP_RECV_BUF, "did not receive buffer")


#define RECV_ERRORS \
	PARA_ERROR(RECV, RECV_SELECT, "recv select error"), \
	PARA_ERROR(RECV, WRITE_STDOUT, "stdout write error"), \
	PARA_ERROR(RECV, HOST_INFO, "stdout write error"), \


#define RECV_COMMON_ERRORS \
	PARA_ERROR(ORTP_RECV, RECV_SYNTAX, "recv syntax error"), \


#define AUDIOD_ERRORS \
	PARA_ERROR(AUDIOD, WRITE_AUDIO_DATA, "failed to write audio data"), \
	PARA_ERROR(AUDIOD, NO_MORE_SLOTS, "no more empty slots"), \
	PARA_ERROR(AUDIOD, MISSING_COLON, "syntax error: missing colon"), \
	PARA_ERROR(AUDIOD, UNSUPPORTED_AUDIO_FORMAT, "given audio format not supported"), \
	PARA_ERROR(AUDIOD, CLIENT_WRITE, "client write error"), \
	PARA_ERROR(AUDIOD, UCRED_PERM, "permission denied"), \
	PARA_ERROR(AUDIOD, INVALID_AUDIOD_CMD, "invalid command"), \


#define FILTER_CHAIN_ERRORS \
	PARA_ERROR(FILTER_CHAIN, UNSUPPORTED_FILTER, "given filter not supported"), \
	PARA_ERROR(FILTER_CHAIN, BAD_FILTER_OPTIONS, "invalid filter option given"), \


#define STAT_ERRORS \
	PARA_ERROR(STAT, TOO_MANY_CLIENTS, "maximal number of stat clients exceeded"), \
	PARA_ERROR(STAT, UNKNOWN_STAT_ITEM, "status item not recognized"), \


#define OGGDEC_ERRORS \
	PARA_ERROR(OGGDEC, OGGDEC_READ, "read from media returned an error"), \
	PARA_ERROR(OGGDEC, OGGDEC_NOTVORBIS, "bitstream is not vorbis data"), \
	PARA_ERROR(OGGDEC, OGGDEC_VERSION, "vorbis version mismatch"), \
	PARA_ERROR(OGGDEC, OGGDEC_BADHEADER, "invalid vorbis bitstream header"), \
	PARA_ERROR(OGGDEC, OGGDEC_FAULT, "bug or heap/stack corruption"), \
	PARA_ERROR(OGGDEC, OGGDEC_BADLINK, "invalid stream section or requested link corrupt"), \


#define GRAB_CLIENT_ERRORS \
	PARA_ERROR(GRAB_CLIENT, GC_AUDIO_FORMAT, "grab client requests unsupported audio format"), \
	PARA_ERROR(GRAB_CLIENT, PEDANTIC_GRAB, "fd not ready and pedantic grab requested"), \
	PARA_ERROR(GRAB_CLIENT, GC_WRITE, "grab client write error"), \
	PARA_ERROR(GRAB_CLIENT, INVALID_GRAB_MODE, "invalid grab client mode"), \
	PARA_ERROR(GRAB_CLIENT, BAD_GC_SLOT, "invalid slot requested by grab client"), \
	PARA_ERROR(GRAB_CLIENT, BAD_GC_FILTER_NUM, "invalid filter number given"), \
	PARA_ERROR(GRAB_CLIENT, GC_SYNTAX, "grab client syntax error"), \
	PARA_ERROR(GRAB_CLIENT, GC_HELP_GIVEN, ""), /* not really an error */ \


#define MP3DEC_ERRORS \
	PARA_ERROR(MP3DEC, MAD_FRAME_DECODE, "mad frame decode error"), \
	PARA_ERROR(MP3DEC, MP3DEC_OVERRUN, "mp3 output buffer overrun"), \


#define FILTER_ERRORS \
	PARA_ERROR(FILTER, NO_FILTERS, "at least one filter must be given"), \
	PARA_ERROR(FILTER, FILTER_SYNTAX, "syntax error"), \


#define SIGNAL_ERRORS \
	PARA_ERROR(SIGNAL, SIGNAL_PIPE, "signal pipe init error"), \
	PARA_ERROR(SIGNAL, SIGNAL_SIG_ERR, "signal() retured SIG_ERR"), \
	PARA_ERROR(SIGNAL, SIGNAL_READ, "read error from signal pipe"), \


#define STRING_ERRORS \
	PARA_ERROR(STRING, MKSTEMP, "mkstemp error: unable to create tmp file"), \
	PARA_ERROR(STRING, FCHMOD, "fchmod error: can not set mode"), \


#define EXEC_ERRORS \
	PARA_ERROR(EXEC, DUP_PIPE, "exec error: can not create pipe"), \
	PARA_ERROR(EXEC, NULL_OPEN, "can not open /dev/null"), \

/* these do not need error handling */
#define WAV_ERRORS
#define COMPRESS_ERRORS
#define TIME_ERRORS
#define CLOSE_ON_FORK_ERRORS
#define DAEMON_ERRORS

extern const char **para_errlist[];

/* a little cpp-magic */
#define SS_SHIFT 8
#define SS_OFFSET(ss) (SS_ ## ss << SS_SHIFT)
#define SS_ENUM(ss) enum {\
	E_ ## ss ## _DUMMY = SS_OFFSET(ss), \
	ss ## _ERRORS}

#define ERRNUM_TO_SS(num) ((num) >> SS_SHIFT)
#define ERRNUM_TO_INDEX(num) (((1 << SS_SHIFT) - 1) & ((num) - 1))
#define PARA_STRERROR(num) para_errlist[ERRNUM_TO_SS(num)] [ERRNUM_TO_INDEX(num)]
#define ERRMSG_ARRAY(ss) [SS_ ## ss] = {ss ## _ERRORS}
#define DEFINE_ERRLIST(ss) const char * ss ## _ERRLIST[] = {ss ## _ERRORS}
#define PARA_ERRLIST(ss) [SS_ ## ss] = ss ## _ERRLIST

/* popcorn time */
#define PARA_ERROR(ss, err, msg) E_ ## err
SS_ENUM(ORTP_RECV);
SS_ENUM(NET);
SS_ENUM(RECV);
SS_ENUM(AUDIOD);
SS_ENUM(EXEC);
SS_ENUM(CLOSE_ON_FORK);
SS_ENUM(SIGNAL);
SS_ENUM(STRING);
SS_ENUM(DAEMON);
SS_ENUM(STAT);
SS_ENUM(TIME);
SS_ENUM(GRAB_CLIENT);
SS_ENUM(HTTP_RECV);
SS_ENUM(RECV_COMMON);
SS_ENUM(FILTER_CHAIN);
SS_ENUM(WAV);
SS_ENUM(COMPRESS);
SS_ENUM(OGGDEC);
SS_ENUM(MP3DEC);
SS_ENUM(FILTER);

/* rest of the world only sees the error text */
#undef PARA_ERROR
#define PARA_ERROR(ss, err, msg) msg
